/* GStreamer
 * Copyright (C) 2009- Luo Jinghua <sunmoon1997@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more
 */

/**
 * SECTION:element-ppssrc
 *
 * This plugin reads data from a remote location specified by a URI.
 * Supported protocols is 'pps'.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>             /* atoi() */
#endif
#include <unistd.h>
#include <gst/gstelement.h>
//#include <gst/gst-i18n-plugin.h>
#include "gstppssrc.h"

GST_DEBUG_CATEGORY_STATIC (ppssrc_debug);
#define GST_CAT_DEFAULT ppssrc_debug

static const GstElementDetails gst_pps_src_details =
GST_ELEMENT_DETAILS ("PPStream source",
    "Source/Network",
    "Receive data as a client over the network via pps",
    "Luo Jinghua <sunmoon1997@gmail.com>");

static GstStaticPadTemplate srctemplate = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS_ANY);

enum
{
  PROP_0,
  PROP_LOCATION,
  PROP_IS_LIVE,
  PROP_OEM,
  PROP_TERMINAL_TYPE,
  PROP_DEVICE_ID,
  PROP_VALID_CODE,
  PROP_VIP_FLAG
};

#define DEFAULT_OEM           "ppstream"
#define DEFAULT_TERMINAL_TYPE  "demo"
#define DEFAULT_DEVICE_ID     "ppsdemo001"

static void gst_pps_src_uri_handler_init (gpointer g_iface,
    gpointer iface_data);
static void gst_pps_src_init (GstPpsSrc * src,
    GstPpsSrcClass * g_class);
static void gst_pps_src_dispose (GObject * gobject);

static void gst_pps_src_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_pps_src_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_pps_src_check_get_range (GstBaseSrc *src);

static GstFlowReturn gst_pps_src_create (GstBaseSrc * src, guint64 offset, guint length,
    GstBuffer ** buffer);

static gboolean gst_pps_src_start (GstBaseSrc * bsrc);

static gboolean gst_pps_src_stop (GstBaseSrc * bsrc);

static gboolean gst_pps_src_get_size (GstBaseSrc * bsrc, guint64 * size);

static gboolean gst_pps_src_is_seekable (GstBaseSrc * bsrc);

static gboolean gst_pps_src_do_seek (GstBaseSrc * bsrc,
    GstSegment * segment);
static gboolean gst_pps_src_unlock (GstBaseSrc * bsrc);

static gboolean gst_pps_src_unlock_stop (GstBaseSrc * bsrc);

static gboolean gst_pps_src_set_location (GstPpsSrc * src,
    const gchar * uri);

static void
_do_init (GType type)
{
  static const GInterfaceInfo urihandler_info = {
    gst_pps_src_uri_handler_init,
    NULL,
    NULL
  };

  g_type_add_interface_static (type, GST_TYPE_URI_HANDLER, &urihandler_info);

  GST_DEBUG_CATEGORY_INIT (ppssrc_debug, "ppssrc", 0,
      "PPStream src");
}

GST_BOILERPLATE_FULL (GstPpsSrc, gst_pps_src, GstBaseSrc,
    GST_TYPE_BASE_SRC, _do_init);

static void
gst_pps_src_base_init (gpointer g_class)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (g_class);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&srctemplate));

  gst_element_class_set_details (element_class, &gst_pps_src_details);
}

static void
gst_pps_src_class_init (GstPpsSrcClass * klass)
{
  GObjectClass *gobject_class;

  GstBaseSrcClass *gstbasesrc_class;

  gobject_class = (GObjectClass *) klass;
  gstbasesrc_class = (GstBaseSrcClass *) klass;

  gobject_class->set_property = gst_pps_src_set_property;
  gobject_class->get_property = gst_pps_src_get_property;
  gobject_class->dispose = gst_pps_src_dispose;

  g_object_class_install_property (gobject_class,
      PROP_LOCATION,
      g_param_spec_string ("location", "Location",
	  "Location to read from", "", G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
      PROP_OEM,
      g_param_spec_string ("oem", "OEM",
          "Value of the OEM pass to PPStream",
          DEFAULT_OEM, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
      PROP_TERMINAL_TYPE,
      g_param_spec_string ("terminal-type", "Terminal type",
          "The type of terminal", DEFAULT_TERMINAL_TYPE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
      PROP_DEVICE_ID,
      g_param_spec_string ("device-id", "device-id",
          "The identify of terminal device", DEFAULT_DEVICE_ID,
          G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
      PROP_VALID_CODE,
      g_param_spec_string ("valid-code", "Valid Code",
          "The valid code", "",
          G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
      PROP_VIP_FLAG,
      g_param_spec_uint ("vip-flag", "VIP flag",
	  "VIP flag", 0, G_MAXUINT32, 0,
          G_PARAM_READWRITE));

  gstbasesrc_class->start = GST_DEBUG_FUNCPTR (gst_pps_src_start);
  gstbasesrc_class->stop = GST_DEBUG_FUNCPTR (gst_pps_src_stop);
  gstbasesrc_class->unlock = GST_DEBUG_FUNCPTR (gst_pps_src_unlock);
  gstbasesrc_class->unlock_stop =
      GST_DEBUG_FUNCPTR (gst_pps_src_unlock_stop);
  gstbasesrc_class->get_size = GST_DEBUG_FUNCPTR (gst_pps_src_get_size);
  gstbasesrc_class->is_seekable =
      GST_DEBUG_FUNCPTR (gst_pps_src_is_seekable);
  gstbasesrc_class->do_seek = GST_DEBUG_FUNCPTR (gst_pps_src_do_seek);

  gstbasesrc_class->check_get_range = GST_DEBUG_FUNCPTR (gst_pps_src_check_get_range);
  gstbasesrc_class->create = GST_DEBUG_FUNCPTR (gst_pps_src_create);

  GST_DEBUG_CATEGORY_INIT (ppssrc_debug, "ppssrc", 0,
      "PPStream Source");
}

static void
gst_pps_src_init (GstPpsSrc * src, GstPpsSrcClass * g_class)
{
  src->location = NULL;
  src->oem = g_strdup (DEFAULT_OEM);
  src->terminal_type = g_strdup (DEFAULT_TERMINAL_TYPE);
  src->device_id = g_strdup (DEFAULT_DEVICE_ID);
  src->valid_code = g_strdup ("");
  src->vip_flag = 0;
  src->interrupted = FALSE;
  src->retry = FALSE;
  src->have_size = FALSE;
  src->seekable = TRUE;
  src->read_position = 0;
  src->request_position = 0;
  src->have_started = FALSE;
}

static void
gst_pps_src_dispose (GObject * gobject)
{
  GstPpsSrc *src = GST_PPS_SRC (gobject);

  GST_DEBUG_OBJECT (src, "dispose");
  g_free (src->location);
  src->location = NULL;
  g_free (src->oem);
  src->oem = NULL;
  g_free (src->terminal_type);
  src->terminal_type = NULL;
  g_free (src->device_id);
  src->device_id = NULL;
  src->fd = -1;

  G_OBJECT_CLASS (parent_class)->dispose (gobject);
}

static void
gst_pps_src_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstPpsSrc *src = GST_PPS_SRC (object);

  switch (prop_id) {
    case PROP_LOCATION:
    {
      const gchar *location;

      location = g_value_get_string (value);

      if (location == NULL) {
        GST_WARNING ("location property cannot be NULL");
        goto done;
      }
      if (!gst_pps_src_set_location (src, location)) {
        GST_WARNING ("badly formatted location");
        goto done;
      }
      break;
    }
    case PROP_OEM:
      if (src->oem)
        g_free (src->oem);
      src->oem = g_value_dup_string (value);
      break;
    case PROP_TERMINAL_TYPE:
      if (src->terminal_type)
        g_free (src->terminal_type);
      src->terminal_type = g_value_dup_string (value);
      break;
    case PROP_DEVICE_ID:
      if (src->device_id)
	g_free (src->device_id);
      src->device_id = g_value_dup_string (value);
      break;
    case PROP_VALID_CODE:
      if (src->valid_code)
	g_free (src->valid_code);
      src->valid_code = g_value_dup_string (value);
      break;
    case PROP_VIP_FLAG:
      src->vip_flag = g_value_get_uint (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
done:
  return;
}

static void
gst_pps_src_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstPpsSrc *src = GST_PPS_SRC (object);

  switch (prop_id) {
    case PROP_LOCATION:
      g_value_set_string (value, src->location);
      break;
    case PROP_OEM:
      g_value_set_string (value, src->oem);
      break;
    case PROP_TERMINAL_TYPE:
      g_value_set_string (value, src->terminal_type);
      break;
    case PROP_DEVICE_ID:
      g_value_set_string (value, src->device_id);
      break;
    case PROP_VALID_CODE:
      g_value_set_string (value, src->valid_code);
      break;
    case PROP_VIP_FLAG:
      g_value_set_uint (value, src->vip_flag);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static gboolean
gst_pps_src_check_get_range (GstBaseSrc *src)
{
  return TRUE;
}

static GstFlowReturn
gst_pps_src_create (GstBaseSrc * psrc, guint64 offset, guint length,
    GstBuffer ** buffer)
{
  int ret;
  int error;
  GstBuffer *buf;
  GstPpsSrc *src = GST_PPS_SRC (psrc);

  if (G_UNLIKELY (!src->have_started)) {
    ret = ppsvod_play_item (src->fd, strlen (src->valid_code) ? src->valid_code : NULL);
    if (ret != 0)
      goto start_failed;
    src->have_started = TRUE;
    src->content_size = ppsvod_get_file_length (src->fd);
    gst_segment_set_duration (&psrc->segment, GST_FORMAT_BYTES, src->content_size);
  }

  if (G_UNLIKELY (src->read_position != offset)) {
    int res;

    res = ppsvod_seek (src->fd, offset);
    if (G_UNLIKELY (res < 0))
      goto seek_failed;

    src->read_position = offset;
  }

  buf = gst_buffer_try_new_and_alloc (length);
  if (G_UNLIKELY (buf == NULL && length > 0)) {
    GST_ERROR_OBJECT (src, "Failed to allocate %u bytes", length);
    return GST_FLOW_ERROR;
  }

  GST_LOG_OBJECT (src, "Reading %d bytes at offset 0x%" G_GINT64_MODIFIER "x",
      length, offset);

  ret = 0;
  while (!src->interrupted && src->fd >= 0) {
    struct timeval tv;

    tv.tv_sec = 0;
    tv.tv_usec = 50000;
    ret = ppsvod_read(src->fd, (char*)GST_BUFFER_DATA (buf),
		      GST_BUFFER_SIZE (buf),
		      src->read_position, &tv);
    error = errno;
    if (ret >= 0)
      break;
    if (error != EAGAIN)
      goto could_not_read;
    usleep (5000);
  }

  if (ret == 0 || src->read_position == src->content_size || src->fd < 0)
    goto eos;

  length = ret;

  GST_BUFFER_SIZE (buf) = length;
  GST_BUFFER_OFFSET (buf) = offset;
  GST_BUFFER_OFFSET_END (buf) = offset + length;

  *buffer = buf;

  src->read_position += length;

  return GST_FLOW_OK;

  /* ERROR */
start_failed:
seek_failed:
  {
    GST_ELEMENT_ERROR (src, RESOURCE, READ, (NULL), GST_ERROR_SYSTEM);
    return GST_FLOW_ERROR;
  }
could_not_read:
  {
    GST_ELEMENT_ERROR (src, RESOURCE, READ, (NULL), GST_ERROR_SYSTEM);
    gst_buffer_unref (buf);
    return GST_FLOW_ERROR;
  }
eos:
  {
    GST_DEBUG ("non-regular file hits EOS");
    gst_buffer_unref (buf);
    return GST_FLOW_UNEXPECTED;
  }
}

static void
gst_pps_src_report_progress (GstPpsSrc *src, int percent)
{
  GstMessage *message;

  message = gst_message_new_element (GST_OBJECT_CAST (src),
				     gst_structure_new ("progress", "percent",
							G_TYPE_INT, percent, NULL));
  gst_element_post_message (GST_ELEMENT_CAST (src), message);
}

static GstPpsSrc * ppssrc = NULL;
static int
gst_pps_src_callback (int index, pps_event event)
{
  unsigned int wparam = event.wparam;
  GstPpsSrc * src = ppssrc;

  if (!src)
    return 0;

  switch (event.id)
    {
    case EPPS_SERVER_TIMEOUT:
      GST_ERROR_OBJECT (src, "server connecting timeout");
      break;
    case EPPS_NO_SERVER:
      GST_WARNING_OBJECT(src, "can not find server");
      break;
    case EPPS_SERVER_NO_RESPONSE:
      GST_WARNING_OBJECT(src, "no response from server");
      break;
    case EPPS_SERVER_INVALID:
      GST_WARNING_OBJECT(src, "server invalid");
      break;
    case EPPS_VERSION_LOW:
      GST_WARNING_OBJECT(src, "client version too low");
      break;
    case EPPS_CLIENT_NAME:
      GST_WARNING_OBJECT(src, "client name error");
      break;
    case EPPS_URL_INVALID:
      GST_WARNING_OBJECT(src, "url not found");
      break;
    case EPPS_NO_SERVICE:
      GST_WARNING_OBJECT(src, "service does not exist");
      break;
    case EPPS_OUT_OF_SERVICE:
      GST_WARNING_OBJECT(src, "out of service area");
      break;
    case EPPS_MEDIA_INFO_ERROR:
      GST_WARNING_OBJECT(src, "loading media info error");
      break;
    case EPPS_INDEX_INFO_ERROR:
      GST_WARNING_OBJECT(src, "index info error");
      break;
    case EPPS_NO_INDEX:
      GST_WARNING_OBJECT(src, "no index");
      break;
    case EPPS_NO_MEIDIA:
      GST_WARNING_OBJECT(src, "no media found");
      break;
    case EPPS_MULTI_INSTANCE:
      GST_WARNING_OBJECT(src, "multi instance running");
      break;
    case EPPS_PREPARE_MEDIA_INFO:
      GST_WARNING_OBJECT(src, "preparing media info %d%%", wparam);
      break;
    case EPPS_GETTING_MEDIA_INFO:
      GST_WARNING_OBJECT(src, "requesting media info %d%%", wparam);
      break;
    case EPPS_GETTING_INDEX_INFO:
      GST_WARNING_OBJECT(src, "requesting index info %d%%", wparam);
      break;
    case EPPS_PLAYING:
      GST_WARNING_OBJECT(src, "playing.");
      break;
    case EPPS_BUFFERING:
      GST_WARNING_OBJECT(src, "buffering %d%%...", wparam);
      gst_pps_src_report_progress (src, wparam);
      break;
    case EPPS_CONNECTING:
      GST_WARNING_OBJECT(src, "connecting to server...");
      break;
    case EPPS_MEDIA_READY:
      GST_WARNING_OBJECT(src, "media is ready");
      break;
    case EPPS_PARSING_SERVER:
      GST_WARNING_OBJECT(src, "parsing server info");
      break;
    case EPPS_GET_MEDIA_INFO:
      GST_WARNING_OBJECT(src, "get media info successfully");
      break;
    case EPPS_BUFFERED:
      GST_WARNING_OBJECT(src, "buffing done");
      gst_pps_src_report_progress (src, 100);
      break;
    case EPPS_PREPARE_MEDIA:
      GST_WARNING_OBJECT(src, "prepare media");
      break;
    case EPPS_STORAGE_ERROR:
      GST_WARNING_OBJECT(src, "access storage failed");
      break;
    case EPPS_DATA_TIMEOUT:
      GST_ERROR_OBJECT(src, "Wait data timeout");
      break;
    case EPPS_UPDATING:
      GST_WARNING_OBJECT(src, "updating %d%%...", wparam);
      break;
    case EPPS_UPDATE_OK:
      GST_WARNING_OBJECT(src, "updating successed");
      break;
    case EPPS_UPDATE_FAILED:
      GST_ERROR_OBJECT(src, "updating failed");
      break;
    case EPPS_QUIT:
      GST_DEBUG_OBJECT(src, "Quit progress");
      break;
    default:
      GST_ERROR_OBJECT(src, "unknown event");
      break;
    }
  return 0;
}

static gboolean
gst_pps_src_start (GstBaseSrc * bsrc)
{
  GstPpsSrc *src = GST_PPS_SRC (bsrc);
  int ret;

  GST_DEBUG_OBJECT (src, "start(\"%s\")", src->location);

  if (!src->location) {
    GST_ELEMENT_ERROR (src, RESOURCE, OPEN_READ,
		       (NULL), ("Missing location property"));
    return FALSE;
  }

  ppssrc = src;
  ret = ppsvod_create (src->oem, src->terminal_type, src->device_id,
		       gst_pps_src_callback);
  if (ret < 0) {
    GST_ELEMENT_ERROR (src, RESOURCE, OPEN_READ,
		       (NULL), ("Couldn't initialize ppsvod"));
    ppssrc = NULL;
    return FALSE;
  }

  GST_DEBUG_OBJECT (src, "add item: (\"%s\")", src->location);
  src->fd = ret = ppsvod_add_item (src->location, src->vip_flag);
  if (ret < 0) {
    GST_ELEMENT_ERROR (src, RESOURCE, OPEN_READ,
		       (NULL), ("Couldn't open resource"));
    ppsvod_destroy();
    ppssrc = NULL;
    return FALSE;
  }

  return TRUE;
}

static gboolean
gst_pps_src_stop (GstBaseSrc * bsrc)
{
  GstPpsSrc *src;

  src = GST_PPS_SRC (bsrc);
  GST_DEBUG_OBJECT (src, "stop()");

  if (src->have_started)
    ppsvod_stop_item (src->fd);
  ppsvod_remove_item (src->fd);
  ppsvod_destroy();

  ppssrc = NULL;
  src->fd = -1;
  src->have_started = FALSE;

  return TRUE;
}

/* Interrupt a blocking request. */
static gboolean
gst_pps_src_unlock (GstBaseSrc * bsrc)
{
  GstPpsSrc *src;

  src = GST_PPS_SRC (bsrc);
  GST_DEBUG_OBJECT (src, "unlock()");

  src->interrupted = TRUE;
  return TRUE;
}

/* Interrupt interrupt. */
static gboolean
gst_pps_src_unlock_stop (GstBaseSrc * bsrc)
{
  GstPpsSrc *src;

  src = GST_PPS_SRC (bsrc);
  GST_DEBUG_OBJECT (src, "unlock_stop()");

  src->interrupted = FALSE;
  return TRUE;
}

static gboolean
gst_pps_src_get_size (GstBaseSrc * bsrc, guint64 * size)
{
  GstPpsSrc *src;

  src = GST_PPS_SRC (bsrc);

  if (!src->have_started)
    return FALSE;

  if (!src->content_size && src->fd >= 0) {
    int ret = ppsvod_get_file_length(src->fd);
    if (ret == -1)
      src->content_size = 0;
    else
      src->content_size = ret;
    GST_DEBUG_OBJECT (src, "get_size() = %" G_GUINT64_FORMAT,
		      src->content_size);
    *size = src->content_size;
    return TRUE;
  } else if (src->content_size) {
    *size = src->content_size;
    return TRUE;
  }

  GST_DEBUG_OBJECT (src, "get_size() = FALSE");
  return FALSE;
}

static gboolean
gst_pps_src_is_seekable (GstBaseSrc * bsrc)
{
  GstPpsSrc *src = GST_PPS_SRC (bsrc);

  GST_DEBUG_OBJECT (src, "seakable: %d", src->seekable);
  return src->seekable;
}

static gboolean
gst_pps_src_do_seek (GstBaseSrc * bsrc, GstSegment * segment)
{
  GstPpsSrc *src = GST_PPS_SRC (bsrc);
  int ret;

  GST_DEBUG_OBJECT (src, "do_seek(%" G_GUINT64_FORMAT ")", segment->start);

  if (src->read_position == segment->start)
    return TRUE;

  if (!src->seekable)
    return FALSE;

  /* Wait for create() to handle the jump in offset. */
  src->request_position = segment->start;
  src->read_position = src->request_position;

  ret = ppsvod_seek (src->fd, segment->start);
  if (ret != 0)
    return FALSE;

  segment->last_stop = segment->start;
  segment->time = segment->start;

  return TRUE;
}

static gboolean
gst_pps_src_set_location (GstPpsSrc * src, const gchar * uri)
{
  if (src->location) {
    g_free (src->location);
    src->location = NULL;
  }
  src->location = g_strdup (uri);

  return TRUE;
}

static guint
gst_pps_src_uri_get_type (void)
{
  return GST_URI_SRC;
}

static gchar **
gst_pps_src_uri_get_protocols (void)
{
  static gchar *protocols[] = { "pps", "tvod", NULL };
  return protocols;
}

static const gchar *
gst_pps_src_uri_get_uri (GstURIHandler * handler)
{
  GstPpsSrc *src = GST_PPS_SRC (handler);

  return src->location;
}

static gboolean
gst_pps_src_uri_set_uri (GstURIHandler * handler, const gchar * uri)
{
  GstPpsSrc *src = GST_PPS_SRC (handler);

  return gst_pps_src_set_location (src, uri);
}

static void
gst_pps_src_uri_handler_init (gpointer g_iface, gpointer iface_data)
{
  GstURIHandlerInterface *iface = (GstURIHandlerInterface *) g_iface;

  iface->get_type = gst_pps_src_uri_get_type;
  iface->get_protocols = gst_pps_src_uri_get_protocols;
  iface->get_uri = gst_pps_src_uri_get_uri;
  iface->set_uri = gst_pps_src_uri_set_uri;
}

static gboolean
plugin_init (GstPlugin * plugin)
{
  return gst_element_register (plugin, "ppssrc", GST_RANK_PRIMARY,
      GST_TYPE_PPS_SRC);
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "pps",
    "PPStream src",
    plugin_init, VERSION, "LGPL", PACKAGE, "")
